/* Author: Denis Podgurskii */

/* Utils */
export class ptk_utils {
    constructor() { }

    static jsonSetValueByPath(jsonData, path, value) {
        if (!(jsonData instanceof Object) || typeof (path) === "undefined") {
            throw "Not valid argument:jsonData:" + jsonData + ", path:" + path
        }
        let origData = jsonData
        path = path.replace(/\[(\w+)\]/g, '.$1') // convert indexes to properties
        path = path.replace(/^\./, '');// strip a leading dot
        let pathArray = path.split('.')
        let i = 0
        do {
            let key = pathArray[i]
            if (key in jsonData) {
                if (i < (pathArray.length - 1)) jsonData = jsonData[key]
                else jsonData[key] = value
            }
            i++
        } while (i < pathArray.length)
        return origData
    }

    static jsonGetValueByPath(jsonData, path) {
        if (!(jsonData instanceof Object) || typeof (path) === "undefined") {
            throw "Not valid argument:jsonData:" + jsonData + ", path:" + path
        }
        let origData = jsonData
        path = path.replace(/\[(\w+)\]/g, '.$1') // convert indexes to properties
        path = path.replace(/^\./, '');// strip a leading dot
        let pathArray = path.split('.')
        let i = 0
        do {
            let key = pathArray[i]
            if (key in jsonData) {
                if (i < (pathArray.length - 1)) jsonData = jsonData[key]
                else return jsonData[key]
            }
            i++
        } while (i < pathArray.length)
        return origData
    }

    static get requestFilters() {
        return ["main_frame", "sub_frame", "stylesheet", "script", "image", "font", "object", "xmlhttprequest", "ping", "csp_report", "media", "websocket", "other"]
    }

    static get extraInfoSpec() {
        return window.isFirefox ? [] : ["extraHeaders"]
    }

    static UUID() {
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
            var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8)
            return v.toString(16)
        })
    }

    static attackId() {
        return 'xxxxxxxxxxxx4xxxyxxxxxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
            var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8)
            return v.toString(16)
        })
    }

    static isURL(url) {
        let regex = new RegExp(/^((http|https):\/\/){1}(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])?(:+[0-9]+)?([\/\?]{1}.*)?$/i)
        return regex.test(url)
    }

    static exclude(url) {
        if (url == 'chrome://newtab/' || url == 'about:newtab') return false
        let regex = new RegExp(/^(chrome:|about:|moz-extension:|chrome-extension:)/i)
        return regex.test(url)
    }
}

/* Logger */
export class ptk_logger {
    constructor() { }

    static log(event, msg, level) {
        if (window.ptk_debug || level == "error") {
            console.log(event)
            if (msg instanceof Array) {
                for (m in msg) {
                    console.log(m + ": " + msg[m])
                }
            } else console.log(msg)
            console.log("Logged at: " + Date.now())
        }
    }

}

/* Notifications -- manage browser notifications */
export class ptk_notifications {
    constructor() { }

    static clearAll() {
        browser.notifications.getAll().then(function (notifications) {
            if (notifications) {
                for (let key in notifications) {
                    browser.notifications.clear(key)
                }
            }
        })
    }

    static notify(title, message, clearAll = true) {
        if (clearAll) this.clearAll()
        browser.notifications.create(
            'PTK_notification', {
            type: 'basic',
            /*iconUrl: browser.extension.getURL('browser/assets/images/icon.png'),*/
            title: title,
            message: message
        })
    }
}

export class ptk_queue {
    constructor() {
        this.items = []
    }

    isEmpty() {
        return (this.items.length === 0)
    }

    enqueue(item) {
        this.items.unshift(item)
    }

    dequeue() {
        return this.items.pop()
    }

    size() {
        return this.items.length
    }

    clear() {
        this.items = []
    }

    has(item) {
        return this.items.includes(item)
    }

}